=begin
 distributed Ruby --- dRuby
 	Copyright (c) 1999-2000 Masatoshi SEKI 
        You can redistribute it and/or modify it under the same term as Ruby.
	$Id: drb.rb,v 1.20 2000/09/19 08:03:04 mas Exp $ 
 DRb --- dRuby module.
 DRbUndumped --- Mixin class. 
 DRbProtocol --- Mixin class.
 DRbObject --- dRuby remote object.
 DRbConn --- 
 DRbIdConv --- id <--> obj converter.
 DRbServer --- dRuby message handler.
=end

require 'socket'

class DRbIdConv
  def to_obj(ref)
    ObjectSpace._id2ref(ref)
  end

  def to_id(obj)
    return nil unless obj
    obj.id 
  end
end

module DRb
  @idconv = DRbIdConv.new
  @acl = nil

  def to_obj(ref)
    @idconv.to_obj(ref)
  end
  def to_id(obj)
    @idconv.to_id(obj)
  end
  module_function :to_id
  module_function :to_obj

  def install_id_conv(f)
    @idconv = f
  end
  module_function :install_id_conv

  def install_acl(acl)
    @acl = acl
  end
  module_function :install_acl

  def start_service(uri=nil, front=nil)
    uri = 'druby://:0' unless uri
    @front = front
    @server = DRbServer.new(uri, @acl)
    @thread = @server.run
  end
  module_function :start_service

  def uri
    @server.uri
  end
  module_function :uri

  attr_reader :thread
  module_function :thread

  attr_reader :front
  module_function :front
end

module DRbUndumped 
  def _dump(dummy)
    raise TypeError, 'can\'t dump'
  end
end

module DRbProtocol
  def parse_uri(uri)
    if uri =~ /^druby:\/\/(.*?):(\d+)(.*)$/
      host = $1
      port = $2.to_i
      option = $3
      [host, port, option]
    else
      raise RuntimeError, 'can\'t parse uri'
    end
  end

  def dump(obj, soc)
    obj = DRbObject.new(obj) if obj.kind_of? DRbUndumped
    begin
      str = Marshal::dump(obj)
    rescue
      ro = DRbObject.new(obj)
      str = Marshal::dump(ro)
    end
    soc.write(str) if soc
    return str
  end

  def send_request(soc, ref,msg_id, *arg, &b)
    dump(ref, soc)
    dump(msg_id.id2name, soc)
    dump(arg.length, soc)
    arg.each do |e|
      dump(e, soc)
    end
    dump(b, soc)
    soc.shutdown(1)
  end
  
  def recv_reply(soc)
    succ = Marshal::load(soc)
    result = Marshal::load(soc)
    [succ, result]
  end

  def recv_request(soc)
    ro = Marshal::load(soc)
    msg = Marshal::load(soc)
    argc = Marshal::load(soc)
    argv = []
    argc.times do
      argv.push Marshal::load(soc)
    end
    block = Marshal::load(soc)
    [ro, msg, argv, block]
  end

  def send_reply(soc, succ, result)
    dump(succ, soc)
    dump(result, soc)
    soc.shutdown(1)
  end

  def ro_to_obj(ro)
    return DRb.front unless ro

    ref = ro.__drbref
    return DRb.front unless ref
    
    return DRb.to_obj(ref)
  end
end

class DRbObject
  def initialize(obj, uri=nil)
    @uri = uri || DRb.uri
    @ref = obj ? DRb.to_id(obj) : nil
  end

  def __drbref
    @ref
  end

  def to_s
    method_missing(:to_s)
  end

  def to_a
    method_missing(:to_a)
  end

  def method_missing(msg_id, *a, &b)
    if @uri == DRb.uri
      if @ref
	obj = DRb.to_obj(@ref)
      else
	obj = DRb.front
      end
      return obj.__send__(msg_id, *a) &b
    end

    succ, result = DRbConn.new(@uri).send_message(self, msg_id, *a, &b)
    raise result if ! succ
    result
  end
end

class DRbConn
  include DRbProtocol

  def initialize(remote_uri)
    @host, @port, opt = parse_uri(remote_uri)
  end

  def send_message(ref, msg_id, *arg, &block)
    begin
      soc = TCPSocket.open(@host, @port)
      send_request(soc, ref, msg_id, *arg, &block)
      recv_reply(soc)
    ensure
      soc.close if soc
#      ObjectSpace.garbage_collect
    end
  end
end

class DRbServer
  include DRbProtocol

  def initialize(uri, acl=nil)
    host, port, opt = parse_uri(uri)
    @soc = TCPServer.open(port)
    host = Socket.gethostname if host.size == 0
    port = @soc.addr[1] if @soc
    @uri = "druby://#{host}:#{port}"
    @acl = acl
  end
  attr_reader(:uri)

  def run
    Thread.start do
      begin
	while true
	  proc
	end
      ensure
	@soc.close if @soc
      end
    end
  end

  def allow?(s)
    return true unless @acl
    return @acl.allow_socket?(s)
  end

  def proc
    ns = @soc.accept
    Thread.start do
      begin
	s = ns	# ??
	Thread.current['DRb'] = { 'socket' => s }
	begin 
	  raise 'Forbidden' unless self.allow?(s)
	  ro, msg, argv, block = recv_request(s)
	  obj = ro_to_obj(ro)
	  if block
	    catch_retry = nil
	    result = obj.__send__(msg.intern, *argv) do |*x|
	      catch_retry = false
	      begin
		block_value = block.call(*x)
	      rescue LocalJumpError
                if $!.message =~ /^retry/	# retry from proc-closure
                  catch_retry = true
                elsif $!.message =~ /^break/	# break from proc-closure
                  break
		else
		  raise $!
                end
	      end
	      retry if catch_retry
	      block_value
	    end
	  else
	    result = obj.__send__(msg.intern, *argv)
	  end
	  succ = true
	rescue
	  result = $!
	  if $VERBOSE 
	    p $!
	    $@.each do |x| puts x end
	  end
	  succ = false
	end
	send_reply(s, succ, result)
      ensure
	s.close if s
#	ObjectSpace.garbage_collect
      end
    end
  end
end

